/*
 * ECRTS 2010 submission "Sulphur"
 * Fully associative cache model
 * Copyright (C) Jack Whitham 2010
 */

#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <assert.h>

#include "cache.h"


typedef struct Cache_Entry_struct {
    int                 is_dirty;
    ulongint            tag, time;
} Cache_Entry;

typedef struct Cache_Params_Struct {
    Tetra_Params *      tp;
    Cache_Entry *       line;
    ulongint            tag_offset;
    ulongint            num_lines;
    ulongint            line_size;
    ulongint            clock;
} Cache_Params;



static int Cache_Worth_Copying (void * cb, ulongint size, 
        ulongint access_count, int read_only)
{
    return 0;
}

static ulongint Cache_Close (void * cb, ulongint value)
{
    return value;
}

static ulongint Cache_Open (void * cb, ulongint value,
        ulongint size, ulongint * offset, int read_only)
{
    return value;
}

static int Cache_Entry_Compare (const void * a, const void * b)
{
    const Cache_Entry * ea = (const Cache_Entry *) a;
    const Cache_Entry * eb = (const Cache_Entry *) b;
    if (ea->tag < eb->tag) {
        return -1;
    } else if (ea->tag > eb->tag) {
        return 1;
    } else {
        return 0;
    }
}

static void Access (void * cb, ulongint value, const char * comment, int write)
{
    Cache_Params *  cp = (Cache_Params *) cb;
    Tetra_Params *  tp = cp->tp;
    Cache_Entry *   te;
    Cache_Entry *   hit;
    Cache_Entry     match;
    ulongint        oldest_entry_time = cp->clock + 2;
    ulongint        place_at = 0;
    ulongint        remove_from = 0;
    ulongint        i, previous_tag = 0;

    /* Assume that hits are more likely than misses.
     * The cache tag store is kept sorted so that a binary
     * search can be used to detect hits. */
    match.tag = (value & ADDRESS_MASK) >> cp->tag_offset;
    hit = bsearch(&match, cp->line, cp->num_lines, sizeof(Cache_Entry),
                Cache_Entry_Compare);
    if (hit != NULL) {
        /* cache hit */
        hit->time = cp->clock + 1;
        cp->clock ++;
        tp->do_tick_fn(tp, TICK_ON_CHIP_ACCESS);
        if (write) {
            hit->is_dirty = 1;
        }
        return;
    }

    /* Cache miss. */
    for (i = 0; i < cp->num_lines; i++) {
        /* Find element for replacement */
        te = &cp->line[ i ];
        assert (te->tag != match.tag);
        if (te->time < oldest_entry_time) {
            oldest_entry_time = te->time;
            remove_from = i;
        }
        /* Find where the replacement should be placed */
        if (te->tag < match.tag) {
            place_at = i + 1;
        }
    }

    /* Make space in the list */
    if (remove_from < place_at) {
        /* Part of list to be shifted left */
        place_at --;
        memmove(&cp->line[ remove_from ],
                &cp->line[ remove_from + 1 ],
                (place_at - remove_from) * sizeof(Cache_Entry));
    } else if (remove_from > place_at) {
        /* Part of list to be shifted right */
        memmove(&cp->line[ place_at + 1 ],
                &cp->line[ place_at ],
                (remove_from - place_at) * sizeof(Cache_Entry));
    }

    /* Replace that element */
    te = &cp->line[ place_at ];
    assert (te != NULL);
    if (te->is_dirty) {
        /* Writeback */
        tp->do_txfer_fn(tp, cp->line_size, 0);
    }

    /* Readin */
    te->is_dirty = write;
    te->tag = match.tag;
    te->time = cp->clock + 1;
    tp->do_txfer_fn(tp, cp->line_size, 0);
    
    /* Check list is still sorted */
    for (i = 0; i < cp->num_lines; i++) {
        te = &cp->line[ i ];
        assert ((te->tag > previous_tag) || 
            ((te->tag == previous_tag) && (previous_tag == ~0)));
        previous_tag = te->tag;
    }
}

static void Cache_Read (void * cb, ulongint value, const char * comment)
{
    Access(cb, value, comment, 0);
}

static void Cache_Write (void * cb, ulongint value, const char * comment)
{
    Access(cb, value, comment, 1);
}

static void Cache_Destroy (void * cb)
{
    Cache_Params *  cp = (Cache_Params *) cb;
    Tetra_Params *  tp = cp->tp;

    free(cp->line);
    free(cp);
    tp->mem_model_cb = NULL;
}

static void Cache_Flush (void * cb)
{
    Cache_Params *  cp = (Cache_Params *) cb;
    Tetra_Params *  tp = cp->tp;
    ulongint        i;

    for (i = 0; i < cp->num_lines; i ++) {
        if (cp->line[ i ].is_dirty) {
            tp->do_txfer_fn(tp, cp->line_size, 0);
        }
        cp->line[ i ].tag = ~0;
        cp->line[ i ].time = 0;
        cp->line[ i ].is_dirty = 0;
    }
}

void Init_Cache(Tetra_Params * tp, ulongint line_size)
{
    Cache_Params *  cp = calloc(1, sizeof(Cache_Params));
    ulongint        i;

    cp->tp = tp;
    cp->line_size = line_size;
    cp->num_lines = tp->bytes_on_chip / line_size;
    cp->clock = 1;
    cp->tag_offset = 0;
    for (i = 1; i < 20; i ++) {
        if ((1 << i) == cp->line_size) {
            cp->tag_offset = i;
            break;
        }
    }
    /* line_size must be a small power of 2 or the following fails: */
    assert (cp->tag_offset != 0); 

    assert (cp->num_lines != 0); 
    cp->line = calloc(cp->num_lines, sizeof(Cache_Entry));

    tp->close_fn = Cache_Close;
    tp->open_fn = Cache_Open;
    tp->read_fn = Cache_Read;
    tp->write_fn = Cache_Write;
    tp->worth_copying_fn = Cache_Worth_Copying;
    tp->destroy_fn = Cache_Destroy;
    tp->flush_fn = Cache_Flush;
    tp->mem_model_cb = cp;

    Cache_Flush(cp);
}

